#! /usr/bin/env python3

###
# This module defines the internal class ``TermSpeaker`` to "speak" on 
# a terminal.
# 
# info::
#     There are also two classes ``BWStylist`` and ``ColorStylist`` to
#     produce the style wanted.
###


from enum import Enum

from .spk_interface import *


# -------------- #
# -- STYLISTS -- #
# -------------- #

###
# prototype::
#     value    = ; // See Python typing...
#                the style code choosen.
#     codetemp = ; // See Python typing...
#                a template that will be updated with the value of 
#                the style code.
#     normcode = ; // See Python typing...
#                the code for the normal style.
#
# This function is just a basic factorization for coding the stylists.
###

def _colorit(
    value   : str,
    codetemp: str,
    normcode: str,
) -> None:
    if value:
        termcode = codetemp.format(value = value)

    else:
        termcode = normcode

    print(termcode, end = "")


###
# This class "colorizes" easily the terminal outputs.
###

# Source: GNU/Linux Mag. - Hors Série 115

class ColorStylist(Enum):
# See ``spk_interface.ALL_CONTEXTS.``
    normal  : str = ''
    good    : str = '96'
    warning : str = '94'
    critical: str = '95'
    error   : str = '31'

    def colorit(self) -> None:
        _colorit(
            value    = self.value,
            codetemp = '\x1b[1;{value}m',
            normcode = '\x1b[0m'
        )

###
# This class uses a "black and white" style for the terminal outputs.
###

class BWStylist(Enum):
# See ``spk_interface.ALL_CONTEXTS.``
    normal  : str = ''
    good    : str = '1m'
    warning : str = '3m'
    critical: str = '1m'
    error   : str = '3m\033[1m'

    def colorit(self) -> None:
        _colorit(
            value    = self.value,
            codetemp = '\033[{value}',
            normcode = '\033[0m'
        )


# ------------------ #
# -- TERM SPEAKER -- #
# ------------------ #

###
# This class implements methods to print ¨infos on the terminal.
###

class TermSpeaker(AbstractSpeaker):
###
# prototype::
#     style    = _ in spk_interface.ALL_GLOBAL_STYLES; // See Python typing...
#                a global style for the outputs.
#     maxwidth = ; // See Python typing...
#                the maw width expected for hard wrapped contents.
###
    def __init__(
        self,
        style,
        maxwidth: int = 80
    ) -> None:
        super().__init__(
            style    = style,
            maxwidth = maxwidth
        )

        self.stylist = {
            GLOBAL_STYLE_COLOR: ColorStylist,
            GLOBAL_STYLE_BW   : BWStylist
        }[self.global_style]


###
# prototype::
#     repeat = (1) ; // See Python typing...
#              the numebr of empty lines wanted.
#
# This method simply prints ``repeat`` empty lines on the terminal.
###
    def NL(self, repeat: int = 1) -> None:
        print("\n"*(repeat - 1))

###
# prototype::
#     text = ; // See Python typing...
#            a text to print as it on the terminal.
###
    def print(self, text: str) -> None:
        print(text)

###
# prototype::
#     context = _ in spk_interface.CONTEXTS (interface.CONTEXT_NORMAL) ; // See Python typing...
#               a context to format some outputs
#
#     :see: = ``ColorStylist`` and ``BWStylist``.
###
    def style(self, context: str = CONTEXT_NORMAL) -> None:
        getattr(self.stylist, context).colorit()


###
# prototype::
#     text = ; // See Python typing...
#            a text to be hard wrapped.
#     tab  = (""); // See Python typing...
#            a possible tabulation to use for each new line created.
#
#     :return: = ; // See Python typing...
#                a wrapped message of maximal width ``self.maxwidth``.
#
# info::
#     We redefine the method ``hardwrap`` because in a terminal, the
#     hard wrapping consists only to add tabulations "to" each new line.
###
    def hardwrap(
        self,
        text: str,
        tab : str = ""
    ) -> str:
        lines = text.split('\n')

        return f'\n{tab}'.join(lines)




#! /usr/bin/env python3

###
# This module defines the class ``Speaker`` that manages easy-to-use recipes
# and the wanted prints (on the terminal and/or in the log file).
###


from functools import wraps
from pathlib import Path

from mistool.term_use import ALL_FRAMES, withframe

from .log  import *
from .term import *


# -------------- #
# -- DECORATE -- #
# -------------- #

# The zero level item will never be used but it simplifies 
# the coding of the API.
ITEM = [
    f'{" "*(4*i)}{deco}'
    if deco != ' ' else
    ''
    for i, deco in enumerate(" *+-")
]

TAB = [
    4*i
    for i in range(0, 4)
]


ASCII_FRAME = {}

for i in range(1, 3):
    ASCII_FRAME[i] = lambda t: withframe(
        text  = t,
        frame = ALL_FRAMES[f'pyba_title_{i}']
    )


# ------------------ #
# -- FOR RECIPES -- #
# ------------------ #

# -- RECIPES - AUTO CODE - START -- #

FORALL  = "forall"
FORLOG  = "forlog"
FORTERM = "forterm"
NL      = "NL"
PRINT   = "print"
PROBLEM = "problem"
STEP    = "step"
STYLE   = "style"
TITLE   = "title"

ACTIONS_NO_ARG = [
    FORALL,
    FORLOG,
    FORTERM,
    NL,
    STYLE,
]

VAR_CONTEXT   = "context"
VAR_INFO      = "info"
VAR_LEVEL     = "level"
VAR_PB_ID     = "pb_id"
VAR_REPEAT    = "repeat"
VAR_STEP_INFO = "step_info"
VAR_TAB       = "tab"
VAR_TEXT      = "text"
VAR_TITLE     = "title"
VAR_WITH_NL   = "with_NL"

# -- RECIPES - AUTO CODE - END -- #


# ---------------------- #
# -- SILENT DECORATOR -- #
# ---------------------- #

###
# This decorator simplifies the managment of the use of ``silent = False``
# when instanciating the class ``Speaker``.
###

def silent_or_not_deco(method):
    @wraps(method)
    def silent_or_not(self, *args, **kwargs) -> None:
        if self.silent:
            self._current_outputs = []

        else:
            method(self, *args, **kwargs)
    
    return silent_or_not


# ------------- #
# -- SPEAKER -- #
# ------------- #

###
# This class is used to "speak": the ¨infos are printed on the terminal 
# and in a log file.
#
# warning::
#     This class must work in any context of use!
###

class Speaker(AbstractSpeaker):
    OUTPUT_LOG  = "log"
    OUTPUT_TERM = "term"
    ALL_OUTPUTS = [OUTPUT_LOG, OUTPUT_TERM]

###
# prototype::
#     logfile = ; // See Python typing...  
#               the path of the log file.
#     style   = _ in spk_interface.ALL_GLOBAL_STYLES ( GLOBAL_STYLE_BW ) ; // See Python typing... 
#     silent  = ( False ) ; // See Python typing...  
#               ``True`` idnicates to print and store nothing contrary to ``False``
#               (this is useful for short processes showing only warning and co
#               when using the method ``resume`` of the class ``problems.Problems``).
###
    def __init__(
        self,
        logfile : Path,
        style   : str  = GLOBAL_STYLE_BW,
        maxwidth: int  = 80,
        silent  : bool = False,
    ) -> None:
# Here we do not need the use of ``super().__init__()``.
        self._speakers = {
            self.OUTPUT_LOG : LogSpeaker(
                logfile  = logfile,
                style    = style,
                maxwidth = maxwidth,
            ),
            self.OUTPUT_TERM: TermSpeaker(
                style    = style,
                maxwidth = maxwidth,
            ),
        }

        self.nbsteps = {
            out: 0
            for out in self.ALL_OUTPUTS
        }
    
        self.silent = silent

        if self.silent:
            self._current_outputs = []

        else:
            self._current_outputs = self.ALL_OUTPUTS

###
# prototype::
#     :see: = speaker.log.LogSpeaker
#
# This method is just an easy-to-use wrapper producing 
# a new empty log file.
###
    def reset_logfile(self) -> None:
        self._speakers[self.OUTPUT_LOG].reset_logfile()


###
# We use ``getter`` and ``setter`` for the boolean attribute ``silent``.
###
    @property
    def silent(self):
        return self._silent

    @silent.setter
    def silent(self, value: bool) -> None:
        self._silent = value
        self.forall()


###
# This method is to use only for a "LOG FILE" output.
###
    @silent_or_not_deco
    def forlog(self) -> None:
        self._current_outputs = [self.OUTPUT_LOG]

###
# This method is to use only for a "TERM" output.
###
    @silent_or_not_deco
    def forterm(self) -> None:
        self._current_outputs = [self.OUTPUT_TERM]

###
# This method is to use all outputs.
###
    @silent_or_not_deco
    def forall(self) -> None:
        self._current_outputs = self.ALL_OUTPUTS


###
# prototype::
#     repeat = (1) ; // See Python typing...
#              the numebr of empty lines wanted.
#
# This method simply prints ``repeat`` empty new lines in all the 
# ouputs wanted.
###
    def NL(self, repeat: int = 1) -> None:
        for out in self._current_outputs:
            self._speakers[out].NL(repeat)

###
# prototype::
#     text = ; // See Python typing...
#            a text to communicate.
###
    def print(self, text: str) -> None:
        for out in self._current_outputs:
            self._speakers[out].print(text)

###
# prototype::
#     context = _ in spk_interface.ALL_CONTEXTS 
#               (interface.CONTEXT_NORMAL) ; // See Python typing...
#               a context for formatting ¨infos.
###
    def style(self, context: str = CONTEXT_NORMAL) -> None:
        for out in self._current_outputs:
            self._speakers[out].style(context)


###
# prototype::
#     title   = ; // See Python typing...
#               the content of the title.
#     level   = _ in [1,2] (1); // See Python typing...
#               the level of the title.
#     with_NL = (True); // See Python typing...
#               ``True`` asks to add a new line after the title and
#               ``False`` to not do this 
#
# info::
#     ``with_NL`` is used to resume problems found, or to print 
#     the very last time stamps in the log file.
###
    def title(self, 
        title  : str,
        level  : int  = 1,
        with_NL: bool = True,
    ) -> None:
        self.print(ASCII_FRAME[level](title))

        if with_NL:
            self.NL()

###
# prototype::
#     step_info = ; // See Python typing...
#                 one short info.
#     level     = _ in [0..3] (0); // See Python typing...
#                 the level of step indicating where ``0`` is for automatic 
#                 numbered enumerations.
###
    def step(self, 
        step_info: str,
        level    : int = 0,
    ) -> None:
        for out in self._current_outputs:
            item = self._stepitem(
                out   = out,
                level = level
            )

            text = self._speakers[out].hardwrap(
                text = f'{item}{step_info}',
                tab  = " "*len(item)
            )

            self._speakers[out].print(text)

###
# This method simpliy resets to `0` the number of numbered steps.
###
    def _reset_nbstep(self) -> None:
        for out in self._current_outputs:
            self.nbsteps[out] = 0

###
# prototype::
#     out   = ; // See Python typing...
#             the kind of speaker.
#     level = _ in [0..3] (0); // See Python typing...
#             the level of step indicating where ``0`` is for automatic 
#             numbered enumerations.
###
    def _stepitem(
        self,
        out  : str,
        level: int = 0,
    ) -> None:
# Enumeration...
        if level == 0:
            self.nbsteps[out] += 1

            return f'{self.nbsteps[out]}) '

# Basic item
        return f'{ITEM[level]} '


###
# prototype::
#     context = ; // See Python typing...
#               the context of a problem.
#     pb_id   = ; // See Python typing...
#               the number of the problem.
#     message = ; // See Python typing...
#               the message to print.
#     level   = _ in [0..3] (0); // See Python typing...
#               the level of the step indicating the problem.
###
    def problem(
        self,
        context: str,
        pb_id  : int,
        info   : str,
        level  : int = 0
    ) -> None:
        self.style(context)

        self.step(
            step_info = f'[ #{pb_id} ] {context.upper()}: {info}',
            level     = level
        )

        self.style(CONTEXT_NORMAL)


###
# prototype::
#     *args = ;
#             the classical list of args allowed by Python.
# 
# This method allows to indicate recipes to apply suchas to simplify 
# the "speaking". Here is an exemple of use followed by the actions 
# actualy done (some actions have short version expressions).
#
# python::
#     self.speaker.receipe(
#         SPEAKER_FOR_TERM,
#         SPEAKER_NL,
#         (SPEAKER_TITLE, f'MONOREPO "{self.monorepo.name}"'),
#         {VAR_TITLE: "STARTING THE ANALYSIS", 
#          VAR_LEVEL: 2}, # A short version here!
#     )
#
# This says to do the following actions.
#
# python::
#     self.speaker.forterm()
#     self.speaker.NL()
#     self.speaker.title(f'MONOREPO "{self.monorepo.name}"')
#     self.speaker.title(
#         title = "STARTING THE ANALYSIS",
#         level = 2
#     )
#
#
# info::
#     One recipe always start and finishes in an "for all" normal context. 
#     This is not optimal but simplifies the writting of recipes.
###
    def recipe(self, *args) -> None:
# Default "for all" normal context.  
        self.forall()
        self.style()

# In most cases, to call the good action with its good arguments we will use: 
# ``getattr(self, action)(*action_args, **action_kwargs)``.
        for action in args:
# An action with no arg.
            if action in ACTIONS_NO_ARG:
                getattr(self, action)()
                continue

# Just a context.
            elif action in ALL_CONTEXTS:
                action_args   = [action]
                action_kwargs = {}  
                action        = STYLE    

# A "string short version" that is not a context: this will just be printed.
            elif type(action) == str:
                self.print(action)
                continue

# A "dict short version": we have to guess the action.
            elif type(action) == dict:
                action_args   = []
                action_kwargs = action   

                if VAR_TITLE in action:
                    action = TITLE

                elif VAR_STEP_INFO in action:
                    action = STEP

                elif VAR_CONTEXT in action:
                    action = PROBLEM

                else:
                    raise ValueError(
                          "impossible to guess the action with the dict:\n"
                        + repr(action)
                    )

# Actions given with args.
            else:
                action_args   = []
                action_kwargs = {}

                action, *extras = action

# ``extras`` is just on dict.
                if (
                    len(extras) == 1 
                    and 
                    type(extras[0]) == dict
                ):
                    action_kwargs = extras[0]

# ``extras`` is a list of args.
                else:
                    action_args = extras

# End of the so clever analysis :-) .
#
# We can call the good action with the good args.
            getattr(self, action)(*action_args, **action_kwargs)

# Default "for all" normal context.  
        self.forall()
        self.style()



#! /usr/bin/env python3

###
# This module defines constants and the interface like class ``AbstractSpeaker``
# that defines a minimal contract for speakers and also some common methods.
###


from abc import ABCMeta, abstractmethod


# --------------- #
# -- CONSTANTS -- #
# --------------- #

# -- INTERFACE - AUTO CODE - START -- #

CONTEXT_CRITICAL = "critical"
CONTEXT_ERROR    = "error"
CONTEXT_FOCUS    = "focus"
CONTEXT_NORMAL   = "normal"
CONTEXT_WARNING  = "warning"

ALL_CONTEXTS = [
    CONTEXT_CRITICAL,
    CONTEXT_ERROR,
    CONTEXT_FOCUS,
    CONTEXT_NORMAL,
    CONTEXT_WARNING
]

# -- INTERFACE - AUTO CODE - END -- #

GLOBAL_STYLE_BW    = "balck & white"
GLOBAL_STYLE_COLOR = "color"

ALL_GLOBAL_STYLES = [
    GLOBAL_STYLE_BW,
    GLOBAL_STYLE_COLOR,
]


# -------------------------------- #
# -- ABSTRACT / INTERFACE CLASS -- #
# -------------------------------- #

###
# This abtract/interface class defines the common ¨api of the speakers.
###

class AbstractSpeaker(metaclass=ABCMeta):
# Source to have a real interface: 
#     * https://realpython.com/python-interface/#using-abcabcmeta 
    @classmethod
    def __subclasshook__(cls, subclass) -> None:
        goodinterface = all(
            hasattr(subclass, methodname) and 
            callable(getattr(subclass, methodname)) 
            for methodname in [
                'print',
                'NL',
            ]
        )

        return goodinterface

###
# prototype::
#     style    = _ in ALL_GLOBAL_STYLES; // See Python typing...
#                a global style for the output. Internally this style is  
#                stored in the attribut ``global_style``.
#     maxwidth = ; // See Python typing...
#                the maw width expected for hard wrapped contents.
###
    def __init__(
        self,
        style   : str,
        maxwidth: int = 80
    ) -> None:
        assert(style in ALL_GLOBAL_STYLES)

        self.maxwidth     = maxwidth
        self.global_style = style


###
# prototype::
#     text = ; // See Python typing...
#            a text to add as it.
###
    @abstractmethod
    def print(self, text: str,) -> None:
        raise NotImplementedError

###
# prototype::
#     repeat = (1) ; // See Python typing...
#              the numebr of empty lines wanted.
###
    @abstractmethod
    def NL(self, repeat: int = 1) -> None:
        raise NotImplementedError


###
# prototype::
#     context = _ in ALL_CONTEXTS (CONTEXT_NORMAL) ; // See Python typing...
#               a context for formatting ¨infos.
#
#
# info::
#     This method doesn't need to be implemented (some speaker has no style 
#     like the log like ones).
###
    def style(self, context: str = CONTEXT_NORMAL) -> None:
# Help for debuging.
#         print(self.__class__)
        ...


###
# prototype::
#     text = ; // See Python typing...
#            a text to be hard wrapped.
#     tab  = (""); // See Python typing...
#            a possible tabulation to use for each new line created.
#
#     :return: = ; // See Python typing...
#                a wrapped message of maximal width ``self.maxwidth``.
###
    def hardwrap(
        self,
        text: str,
        tab : str = ""
    ) -> str:
        shortlines = []

        for oneline in text.split('\n'):
            words = [w.strip() for w in oneline.split(' ')]

            if shortlines:
                lastline = tab
            
            else:
                lastline = ""

            lastline += words.pop(0)


            while(words):
                oneword = words.pop(0)

                len_lastline = len(lastline)
                len_word     = len(oneword)

                if len_lastline + len_word >= self.maxwidth :
                    shortlines.append(lastline)
                    lastline = tab

                else:
                    lastline += " "
                    
                lastline += oneword

            shortlines.append(lastline)

        return "\n".join(shortlines)



#! /usr/bin/env python3

###
# This module defines the internal class ``LogSpeaker`` to "speak" in 
# a log file.
###


from pathlib import Path

from .spk_interface import *


# ----------------- #
# -- LOG SPEAKER -- #
# ----------------- #

###
# This class implements methods to print ¨infos in the log file.
###

class LogSpeaker(AbstractSpeaker):
###
# prototype::
#     logfile  = ; // See Python typing...  
#                the path of the log file.
#     style    = _ in spk_interface.ALL_GLOBAL_STYLES; // See Python typing...
#                a global style for the outputs.
#     maxwidth = ; // See Python typing...
#                the maw width expected for hard wrapped contents.
###
    def __init__(
        self,
        logfile : Path,
        style   : str,
        maxwidth: int,
    ) -> None:
        super().__init__(
            style    = style,
            maxwidth = maxwidth
        )

        self.logfile = logfile

        self.reset_logfile()

###
# This method produces a new empty log file.
###
    def reset_logfile(self) -> None:
# Empty an existing log file.
        if self.logfile.is_file():
            with self.logfile.open(
                encoding = "utf8",
                mode     = "w"
            ) as logfile:
                logfile.write("")

# New log file if it doesn't exist.
        else:
            self.logfile.touch()


###
# prototype::
#     text = ; // See Python typing...
#            a text to print as it in the log file.
###
    def print(
        self, text  : str,
    ) -> None:
        with self.logfile.open(
            encoding = "utf8",
            mode     = "a"
        ) as logfile:
            logfile.write(text)
            logfile.write("\n")


###
# prototype::
#     repeat = (1) ; // See Python typing...
#              the numebr of empty lines wanted.
#
# This method simply append ``repeat`` empty new lines to the log file.
###
    def NL(self, repeat: int = 1) -> None:
        self.print("\n"*(repeat - 1))



#!/usr/bin/env python3

from datetime import datetime

from .speaker import *


# ---------------- #
# -- TIME STAMP -- #
# ---------------- #

###
# prototype::
#     speaker = speaker.allinone.Speaker ;  
#               the class used to speak in the log file (no time stamp printed
#               in a terminal).
#     kind    = ; // See Python typing...
#               the kind of time stamp ("start" and "end" for example). 
#               This string will be always "upperized".
#     with_NL = (True); // See Python typing...
#               ``True`` asks to add a new line after the title and
#               ``False`` to not do this.
###

def timestamp(
    speaker: Speaker,
    kind   : str,
    with_NL: bool = True,
) -> None:
    now = datetime.now().strftime("%Y-%m-%d (%H:%M:%S)")

    timeTXT = f"{kind.upper()} TIME STAMP: {now}"

    speaker.recipe(
        FORLOG,
            {VAR_TITLE  : timeTXT, 
             VAR_LEVEL  : 2, 
             VAR_WITH_NL: with_NL},
    )



#! /usr/bin/env python3

###
# This module defines the class ``Problems`` to indicate warnings,
# "criticals" and errors.
###


from collections import defaultdict

from natsort import natsorted

from .speaker import *


# ------------------------ #
# -- PROBLEMS DECORATOR -- #
# ------------------------ #

###
# This decorator simplifies the managment of the new warnings, criticals
# and errors.
###

def problems_deco(method):
# what -> any object with a string representation indicating clearly 
#         what is causing the problem.
#
# info -> the info explaining the error.
#
# level -> the level of the step indicating the problem.
    def similar(
        self,
        what : Any,
        info : str,
        level: int = 0
    ) -> None:
        kind = method.__name__.replace('new_', '')

        nb_kind = f'nb_{kind}s'

        lastnb = getattr(self, nb_kind)
        setattr(self, nb_kind, lastnb + 1)

        self._new_pb(
            what    = what,
            context = globals()[f'CONTEXT_{kind.upper()}'],
            info    = info,
            level   = level
        )
    
    return similar


# -------------- #
# -- PROBLEMS -- #
# -------------- #

###
# This class is used to store ¨infos about errors and warnings 
# emitted during all the process.
###

class Problems:
    PB_INFO_TAG: str = 'info'
    PB_ID_TAG  : str = 'pbid'

###
# prototype::
#     speaker = ; // See Python typing...  
#               an instance of ``toolbox.speaker.allinone.Speaker`` 
#               is used to communicate small ¨infos.
###
    def __init__(
        self,
        speaker: Speaker,
    ) -> None:
        self.speaker = speaker

# ---------
# WARNING !
# ---------
# 
# We use the ordered feature of dict such as to treat warnings, criticals and
# errors in this order when printing the summaries.
        self.problems_found : dict = {
            CONTEXT_WARNING : defaultdict(list), # Before ERROR and CRITICAL!
            CONTEXT_CRITICAL: defaultdict(list), # Before ERROR!
            CONTEXT_ERROR   : defaultdict(list),
        }

        self.nb_warnings  = 0
        self.nb_criticals = 0
        self.nb_errors    = 0

        self._pb_id = 0

###
# prototype::
#     :see: = speaker.log.LogSpeaker
#
# This method is just an easy-to-use wrapper producing 
# a new empty log file.
###
    def reset_logfile(self) -> None:
        self.speaker.reset_logfile()

###
# prototype::
#     :return: = ; // See Python typing...
#                ``True`` if at least one warning has been found and
#                ``False` otherwise.
###
    @property
    def warningfound(self) -> bool:
        return self.nb_warnings != 0

###
# prototype::
#     :return: = ; // See Python typing...
#                ``True`` if at least one "critical" has been found and
#                ``False` otherwise.
###
    @property
    def criticalfound(self) -> bool:
        return self.nb_warnings != 0

###
# prototype::
#     :return: = ; // See Python typing...
#                ``True`` if at least one error has been found and
#                ``False` otherwise.
###
    @property
    def errorfound(self) -> bool:
        return self.nb_errors != 0

###
# prototype::
#     :return: = ; // See Python typing...
#                ``True`` if at least on error or one warning has been found and
#                ``False` otherwise.
###
    @property
    def pbfound(self) -> bool:
        return (
            self.warningfound
            or
            self.criticalfound
            or
            self.errorfound
        )

###
# prototype::
#     :return: = ; // See Python typing...
#                ``True`` if there are several warnings and
#                ``False`` otherwise.
###
    @property
    def several_warnings(self) -> bool:
        return self.nb_warnings > 1

###
# prototype::
#     :return: = ; // See Python typing...
#                ``True`` if there are several "criticals" and
#                ``False`` otherwise.
###
    @property
    def several_criticals(self) -> bool:
        return self.nb_criticals > 1

###
# prototype::
#     :return: = ; // See Python typing...
#                ``True`` if there are several erros and
#                ``False`` otherwise.
###
    @property
    def several_errors(self) -> bool:
        return self.nb_errors > 1


###
# prototype::
#     what  = ; // See Python typing...
#             any object with a string representation indicating clearly 
#             what is causing the warning.
#     info  = ; // See Python typing...
#             the info explaining the warning.
#     level = _ in [0..3] (0); // See Python typing...
#             the level of the step indicating the problem.
###
    @problems_deco
    def new_warning(
        self,
        what : Any,
        info : str,
        level: int = 0
    ) -> None:
        ...

###
# prototype::
#     what  = ; // See Python typing...
#             any object with a string representation indicating clearly 
#             what is causing the "critical" which is a dangerous warning.
#     info  = ; // See Python typing...
#             the info explaining the "critical".
#     level = _ in [0..3] (0); // See Python typing...
#             the level of the step indicating the problem.
###
    @problems_deco
    def new_critical(
        self,
        what : Any,
        info : str,
        level: int = 0
    ) -> None:
        ...

###
# prototype::
#     what  = ; // See Python typing...
#             any object with a string representation indicating clearly 
#             what is causing the error.
#     info  = ; // See Python typing...
#             the info explaining the error.
#     level = _ in [0..3] (0); // See Python typing...
#             the level of the step indicating the problem.
###
    @problems_deco
    def new_error(
        self,
        what : Any,
        info : str,
        level: int = 0
    ) -> None:
        ...

###
# prototype::
#     what    = ; // See Python typing...
#               any object with a string representation indicating clearly 
#               what is causing the problem.
#     context = _ in [speaker.spk_interface.CONTEXT_ERROR, 
#                     speaker.spk_interface.CONTEXT_WARNING] ; 
#               the kind of problem.
#     info    = ; // See Python typing...
#               the info explaining the problem.
#     level   = _ in [0..3] (0); // See Python typing...
#               the level of the step indicating the problem.
###
    def _new_pb(
        self,
        what   : Any,
        context: str,
        info   : str,
        level  : int = 2
    ) -> None:
# Let's store the problems internally.
        self._pb_id += 1
        self.problems_found[context][str(what)].append({
            self.PB_ID_TAG  : self._pb_id,
            self.PB_INFO_TAG: info,
        })
    
# Let's talk to the world...
        self.speaker.problem(
            context = context,
            info    = info,
            level   = level,
            pb_id   = self._pb_id
        )


###
# This method prints the summaries for the terminal and the log file.
###
    def resume(self) -> None:
# Silent must be switch off here...
        silent_user         = self.speaker.silent
        self.speaker.silent = False

# Let's talk...
        showref = add_NL = not silent_user

        for context, pbs in self.problems_found.items():
            if not pbs:
                continue

# Header
            total_nb_pbs = getattr(self, f'nb_{context.lower()}s')

            plurial = "S" if total_nb_pbs > 1 else ""

            if add_NL:
                self.speaker.NL()

            if silent_user:
                add_NL = True

            self.speaker.recipe(
                context,
                {VAR_TITLE:
                    f'{total_nb_pbs} {context.upper()}{plurial} FOUND',
                 VAR_LEVEL  : 2,
                 VAR_WITH_NL: False},
            )

# Verbose in the terminal?
            if not silent_user:
                self.speaker.recipe(
                    context,
                    FORTERM,
                        NL,
                        'Look at the log file or above for details.',
                )

# The problems (cardinality + refs).
            for what in natsorted([
                str(p) for p in pbs
            ]):
                if silent_user:
                    verb_in = [FORTERM, FORLOG]

                else:
                    verb_in = [FORLOG]

                    self._resume_one_pb_short(
                        what    = what,
                        whatpbs = pbs[what],
                        context = context,
                    )

                self._resume_one_pb_verbose(
                    what    = what,
                    whatpbs = pbs[what],
                    context = context,
                    verb_in = verb_in,
                    showref = showref,
                )

# Let's go back to the silent user's choice.
        self.speaker.silent = silent_user

###
# prototype::
#     what    = ; // See Python typing...
#               any object with a string representation indicating clearly 
#               what is causing the problem.
#     whatpbs = ; // See Python typing...
#               the list of problems for the same ``what``.
#     context = _ in [speaker.spk_interface.CONTEXT_ERROR, 
#                     speaker.spk_interface.CONTEXT_WARNING] ; 
#               the kind of problem.
###
    def _resume_one_pb_short(
        self,
        what   : str,
        whatpbs: List[dict],
        context: str,
    ):
        pbs_ids = [
            onepb[self.PB_ID_TAG]
            for onepb in whatpbs
        ]

        nb_pbs = len(whatpbs)

        plurial = "s" if nb_pbs > 1 else ""

        self.speaker.recipe(
            FORTERM,
                context,
                NL,
                {VAR_STEP_INFO: what, 
                VAR_LEVEL    : 1},
        )
                
        self.speaker.recipe(
            FORTERM,
                context,
                    {VAR_STEP_INFO: (
                        f'{nb_pbs} {context}{plurial}.'
                        '\n'
                        f'See #.: {pbs_ids}.'), 
                    VAR_LEVEL: 2},
        )


###
# prototype::
#     what    = ; // See Python typing...
#               any object with a string representation indicating clearly 
#               what is causing the problem.
#     whatpbs = ; // See Python typing...
#               the list of problems for the same ``what``.
#     context = _ in [speaker.spk_interface.CONTEXT_ERROR, 
#                     speaker.spk_interface.CONTEXT_WARNING] ; 
#               the kind of problem.
#               the list of problems for the same ``what``.
#     verb_in = ; // See Python typing...
#               the list of outputs where to be verbose.
#     showref = ; // See Python typing...
#               ``True`` asks to show the references (for a none silent mode)
#               contrary to ``False``.
###
    def _resume_one_pb_verbose(
        self,
        what   : str,
        whatpbs: List[dict],
        context: str,
        verb_in: List[str],
        showref: bool,
    ):
        for output in verb_in:
            self.speaker.recipe(
                output,
                    context,
                    NL,
                    {VAR_STEP_INFO: what, 
                     VAR_LEVEL    : 1},
            )
            
            for onepb in whatpbs:
                info = onepb[self.PB_INFO_TAG]

                if showref:
                    pbid = onepb[self.PB_ID_TAG]

                    message = f'See [ #.{pbid} ] : {info}'
                
                else:
                    message = info[0].upper() + info[1:]

                self.speaker.recipe(
                    output,
                        context,
                        {VAR_STEP_INFO: message, 
                        VAR_LEVEL    : 2},
                )



#! /usr/bin/env python3

from typing import *

from pathlib import Path

from .problems import *
from .timer    import *


# ------------------------------ #
# -- BASE COMMUNICATING CLASS -- #
# ------------------------------ #

###
# This class gives the common interface for classes working with "speakers" 
# and "problems".
###

class BaseCom:

###
# prototype::
#     problems = ; // See Python typing...  
#                an instance of ``toolbox.Problems`` that manages 
#                a basic history of the problems found.
#
# info::
#     This class uses the attribut ``success`` to indicate if
#     at least one "critical" or error has been emitted (sometimes
#     we want to continue to work even if something bad has been
#     found, for example when working on several projects in the
#     same monorepo)
###
    def __init__(
        self,
        problems: Problems,
    ) -> None:
        self.problems = problems
        self.success  = True


###
# prototype::
#     :see: = speaker.log.LogSpeaker
#
# This method is just an easy-to-use wrapper producing 
# a new empty log file.
###
    def reset_logfile(self):
        self.problems.reset_logfile()


###
# prototype::
#     :see: = problems.Problems.new_warning
# 
# This method is just an easy-to-use wrapper.
###
    def new_warning(self, *args, **kwargs):
        self.problems.new_warning(*args, **kwargs)

###
# prototype::
#     :see: = problems.Problems.new_warning
# 
# This method is just an easy-to-use wrapper.
#
# info::
#     The difference between a warning and a critical is that a critical is 
#     a warning that blocks one part of the process but not all the process.
#     It is a kind of weak error or very strong warning.
###
    def new_critical(self, *args, **kwargs):
        self.success = False
        self.problems.new_critical(*args, **kwargs)

###
# prototype::
#     :see: = problems.Problems.new_error
# 
# This method is just an easy-to-use wrapper.
###
    def new_error(self, *args, **kwargs):
        self.success = False
        self.problems.new_error(*args, **kwargs)

###
# prototype::
#     :see: = problems.Problems.resume
# 
# This method is just an easy-to-use wrapper.
###
    def resume(self, *args, **kwargs):
        self.problems.resume(*args, **kwargs)


###
# prototype::
#     :see: = speaker.allinone.Speaker.recipe
# 
# This method is just an esay-to-use wrapper.
###
    def recipe(self, *args, **kwargs) -> None:
        self.problems.speaker.recipe(*args, **kwargs)



###
# prototype::
#     kind    = ; // See Python typing...
#               the kind of time stamp ("start" and "end" for example). 
#               This string will be always "upperized".
#     with_NL = (True); // See Python typing...
#               ``True`` asks to add a new line after the title and
#               ``False`` to not do this.
#
#     :see: = timer.timestamp
# 
# This method is just an easy-to-use wrapper.
###
    def timestamp(
        self,
        kind   : str,
        with_NL: bool = True
    ):
        timestamp(
            speaker = self.problems.speaker,
            kind    = kind,
            with_NL = with_NL,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spkpb', 'spkpb.speaker']

package_data = \
{'': ['*']}

install_requires = \
['mistool>=1.2.6-beta.0,<2.0.0', 'natsort>=7.1.1,<8.0.0']

setup_kwargs = {
    'name': 'spkpb',
    'version': '0.0.8b0',
    'description': 'A module to write programs which have to be verbose about a process on files and directories, and that have to emit informations, warnings and errors.',
    'long_description': 'The `Python` module `spkpb`\n===========================\n\n\n> **I beg your pardon for my english...**\n>\n> English is not my native language, so be nice if you notice misunderstandings, misspellings, or grammatical errors in my documents and codes.\n\n\nAbout `spkpb`\n-------------\n\nThis module proposes two classes that simplify the writing of programs which have to be verbose about a process on files and directories, and that have to emit informations, warnings and errors.\n\n  1. `Speaker`, the `spk` of `spkpb`, has methods tho print informations on a terminal and/or in a log file.\n  \n  1. `Problems`, the `pb` of `spkpb`, allows to indicate and store warnings, "criticals" and errors.\n\n\nThere is also a function `timestamp` to add time stamps in a log file without printing anything in the terminal.\n\n\nUsing directly the API - Default mode\n-------------------------------------\n\n### `Python` code\n\nLet\'s consider the following `Python` file where `Path` is a class proposed by the module `pathlib`. You have to know that the values of the arguments ``what`` are "stringified" (this allows to use either standard strings or advanced classes by defining your own ``__str__`` method for the resume output of problems if you need it).\n\n```python\nfrom spkpb import *\n\nspeaker = Speaker(\n    logfile = Path(\'mylog.log\')\n)\n\nproblems = Problems(speaker)\n\nproblems.new_warning(\n    what = Path(\'one/strange/file.txt\'),\n    info = "some strange behaviors."\n)\n\nproblems.new_error(\n    what = Path(\'one/bad/file.txt\'),\n    info = "bad things appear."\n)\n\nspeaker.recipe(\n    NL,\n    \'One basic showcase.\',\n    FORTERM,\n        {VAR_STEP_INFO: \'ONLY FOR THE TERMINAL OUPUT!\',\n         VAR_LEVEL    : 1},\n    FORLOG,\n        {VAR_STEP_INFO: \'ONLY IN THE LOG FILE!\',\n         VAR_LEVEL    : 1},\n)\n    \nproblems.resume()\n```\n\n### The terminal output\n\nLaunching our `Python` code from a terminal, we will see the following output.\n\n```\n1) [ #1 ] WARNING: some strange behaviors.\n2) [ #2 ] ERROR: bad things appear.\n\nOne basic showcase.\n    * ONLY FOR THE TERMINAL OUPUT!\n\n---------------\n1 WARNING FOUND\n---------------\n\nLook at the log file and/or above for details.\n\n    * "one/strange/file.txt"\n        + 1 warning.\n          See #.: [1].\n\n-------------\n1 ERROR FOUND\n-------------\n\nLook at the log file and/or above for details.\n\n    * "one/bad/file.txt"\n        + 1 error.\n          See #.: [2].\n```\n\n\n### The content of the log file `mylog.log`\n\nLaunching our `Python` code, `mylog.log` will have the following content (just note that the resume is more verbose than the one in a terminal).\n\n\n```\n1) [ #1 ] WARNING: some strange behaviors.\n2) [ #2 ] ERROR: bad things appear.\n\nOne basic showcase.\n    * ONLY IN THE LOG FILE!\n\n---------------\n1 WARNING FOUND\n---------------\n\n    * one/strange/file.txt\n        + See [ #.1 ] : some strange behaviors.\n\n-------------\n1 ERROR FOUND\n-------------\n\n    * one/bad/file.txt\n        + See [ #.2 ] : bad things appear.\n```\n\n\nUsing directly the API - Silent mode\n------------------------------------\n\nLet\'s modify a little our first code (the ellipsis indicate lines unchanged).\n\n```python\nfrom spkpb import *\n\nspeaker = Speaker(\n    logfile = Path(\'mylog.log\'),\n    silent  = True\n)\n\n...\n```\n\nThe use of the argument ``silent`` asks to prints only the summaries of problems (that is useful for short processes with no need to be verbose). The terminal and the log file will show the following same verbose resume.\n\n```\n---------------\n1 WARNING FOUND\n---------------\n\n    * one/strange/file.txt\n        + Some strange behaviors.\n\n-------------\n1 ERROR FOUND\n-------------\n\n    * one/bad/file.txt\n        + Bad things appear.\n```\n\n\nTime stamp in the log file\n--------------------------\n\nThe following code show how to use `timestamp` such as to add time stamps in the log file.\n\n```python\nfrom spkpb import *\n\nspeaker = Speaker(\n    logfile = Path(\'mylog.log\')\n)\n\ntimestamp(\n    speaker = speaker,\n    kind    = \'start 1\',\n)\n\ntimestamp(\n    speaker = speaker,\n    kind    = \'start 2\',\n    with_NL = False,\n)\n\ntimestamp(\n    speaker = speaker,\n    kind    = \'start 3\',\n)\n```\n\nThis will add the following lines in the log file `mylog.log` without printing anything in the terminal.\n\n```\n-----------------------------------------\nSTART 1 TIME STAMP: 2021-08-09 (00:40:02)\n-----------------------------------------\n\n-----------------------------------------\nSTART 2 TIME STAMP: 2021-08-09 (00:40:02)\n-----------------------------------------\n-----------------------------------------\nSTART 3 TIME STAMP: 2021-08-09 (00:40:02)\n-----------------------------------------\n\n```\n\n\nA ready-to-use communicating class\n----------------------------------\n\nWe have seen hard use of the API of `spkpb`. Indeed you can heritate the class `BaseCom` to do things easily: see the following code and outputs.\n\n\n### `Python` code\n\n~~~python\nfrom spkpb import *\n\nproject = BaseCom(\n    Problems(\n        Speaker(\n            logfile = Path(\'mylog.log\')\n        )\n    )\n)\n\nproject.timestamp(kind = \'start\')\n\nproject.new_warning(\n    what = Path(\'one/strange/file.txt\'),\n    info = "some strange behaviors."\n)\n\nprint(f\'>>>>>>>> sucess = {project.success}\')\n\nproject.new_error(\n    what = Path(\'one/bad/file.txt\'),\n    info = "bad things appear."\n)\n\nprint(f\'>>>>>>>> sucess = {project.success}\')\n\nproject.recipe(\n    NL,\n    \'One basic showcase.\',\n    FORTERM,\n        {VAR_STEP_INFO: \'ONLY FOR THE TERMINAL OUPUT!\',\n         VAR_LEVEL    : 1},\n    FORLOG,\n        {VAR_STEP_INFO: \'ONLY IN THE LOG FILE!\',\n         VAR_LEVEL    : 1},\n)\n    \nproject.resume()\n\nproject.recipe(NL)\nproject.timestamp(kind = \'end\')\n~~~\n\n\n### The terminal output\n\n~~~\n1) [ #1 ] WARNING: some strange behaviors.\n>>>>>>>> sucess = True\n2) [ #2 ] ERROR: bad things appear.\n>>>>>>>> sucess = False\n\nOne basic showcase.\n    * ONLY FOR THE TERMINAL OUPUT!\n\n---------------\n1 WARNING FOUND\n---------------\n\nLook at the log file or above for details.\n\n    * one/strange/file.txt\n        + 1 warning.\n          See #.: [1].\n\n-------------\n1 ERROR FOUND\n-------------\n\nLook at the log file or above for details.\n\n    * one/bad/file.txt\n        + 1 error.\n          See #.: [2].\n~~~\n\n\n### The content of the log file `mylog.log`\n\n~~~\n---------------------------------------\nSTART TIME STAMP: 2021-08-10 (11:40:02)\n---------------------------------------\n\n1) [ #1 ] WARNING: some strange behaviors.\n2) [ #2 ] ERROR: bad things appear.\n\nOne basic showcase.\n    * ONLY IN THE LOG FILE!\n\n---------------\n1 WARNING FOUND\n---------------\n\n    * one/strange/file.txt\n        + See [ #.1 ] : some strange behaviors.\n\n-------------\n1 ERROR FOUND\n-------------\n\n    * one/bad/file.txt\n        + See [ #.2 ] : bad things appear.\n\n-------------------------------------\nEND TIME STAMP: 2021-08-10 (11:40:02)\n-------------------------------------\n~~~\n\n\nReset the log file\n------------------\n\nThe classes `BaseCom`, `Problems` and `Speaker` all have a method `reset_logfile` to reset the log file. Here is a relly weird example of use even if the method is indeed very useful each time the communicating process restarts.\n\n~~~python\nfrom spkpb import *\n\nproject = BaseCom(\n    Problems(\n        Speaker(\n            logfile = Path(\'mylog.log\')\n        )\n    )\n)\n\nproject.new_warning(\n    what = Path(\'one/strange/file.txt\'),\n    info = "some strange behaviors."\n)\n\nproject.reset_logfile()\n\nproject.new_error(\n    what = Path(\'one/bad/file.txt\'),\n    info = "bad things appear."\n)\n~~~\n\n\nIn that case, the log file will be the following one with the 1st warning missing. Who has chosen this excellent example? :-)\n\n~~~\n2) [ #2 ] ERROR: bad things appear.\n~~~\n\n\n<!-- :tutorial-START: -->\n<!-- :tutorial-END: -->\n\n\n<!-- :version-START: -->\n<!-- :version-END: -->\n',
    'author': 'Christophe BAL',
    'author_email': None,
    'maintainer': 'Christophe BAL',
    'maintainer_email': None,
    'url': 'https://github.com/projetmbc/tools-for-dev/tree/master/spkpb',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
